@@ -4,49 +4,56 @@
 #include <thread>
 #include <cstdio>
 #include <vector>
+#include <mutex>
+#include <regex>
+#include <chrono>
+#include <cmath>
 extern "C" {
 #include "fts-xapian-plugin.h"
 }
 #include <dirent.h>
 #include <unicode/utypes.h>
 #include <unicode/unistr.h>
-#include <unicode/translit.h>
+#include <unicode/translit.h> 
 #include <sys/time.h>
 
-#define HDRS_NB 10
-static const char * hdrs_emails[HDRS_NB] = { "uid", "subject", "from", "to",  "cc",  "bcc",  "messageid", "listid", "body", ""  };
-static const char * hdrs_xapian[HDRS_NB] = { "Q",   "S",       "A",    "XTO", "XCC", "XBCC", "XMID",      "XLIST",  "XBDY", "XBDY" };
-static const char * createTable = "CREATE TABLE IF NOT EXISTS docs(ID INT PRIMARY KEY NOT NULL);";
-static const char * selectUIDs = "select ID from docs;";
-#define CHARS_PB 18
-#define CHAR_KEY "_"
-#define CHAR_SPACE " "
-static const char * chars_pb[] = { "<", ">", ".", "-", "@", ":", ";", "!", "%", "*", "|", "`", "#", "~", "^", ",", "/", "\\" };
+#include <pwd.h>
+#include <grp.h>
+#include <sys/stat.h>
+
+#include <syslog.h>
+
+class XDoc;
+class XDocsWriter;
+
+typedef std::vector<XDoc *> XDocs;
+
+#define XSLEEP std::chrono::milliseconds(200)
 
 struct xapian_fts_backend
 {
 	struct fts_backend backend;
-	char * path = NULL;
+	char * path;
 
 	char * guid;
 	char * boxname;
 	char * db;
 	char * expdb;
+	Xapian::WritableDatabase * dbw;
 
 	char * old_guid;
 	char * old_boxname;
 
-	Xapian::WritableDatabase * dbw;
+        XDocs * docs;
+	long threads_total;
+	std::vector<XDocsWriter *> threads;
+	long threads_max;
+	std::timed_mutex mutex;
 
-	long commit_updates;
-	long commit_time;
-
-	long perf_pt;
-	long perf_nb;
-	long perf_uid;
-	long perf_dt;
-
-	long nbdocs = 0;
+	long lastuid;
+	long total_added_docs;
+	long batch;
+	long start_time;
 };
 
 struct xapian_fts_backend_update_context
@@ -79,8 +86,17 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *)_backend;
 
 	backend->db = NULL;
-	backend->dbw = NULL;
 
+	backend->docs = NULL;
+	backend->threads.clear();
+	backend->threads_max = std::thread::hardware_concurrency();
+        backend->threads_total = 0;
+	
+	backend->lastuid = -1;
+	backend->total_added_docs = 0;
+	backend->batch = 0;
+
+	backend->dbw = NULL;
 	backend->guid = NULL;
 	backend->path = NULL;
 	backend->old_guid = NULL;
@@ -91,7 +107,9 @@ static int fts_backend_xapian_init(struct fts_backend *_backend, const char **er
 
 	if(fts_backend_xapian_set_path(backend)<0) return -1;
 
-        if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Starting with partial=%ld full=%ld verbose=%d lowmemory=%ld MB vs freemem=%ld MB", fts_xapian_settings.partial,fts_xapian_settings.full,fts_xapian_settings.verbose,fts_xapian_settings.lowmemory, long(fts_backend_xapian_get_free_memory()/1024.0));
+	openlog("xapian-docswriter",0,LOG_MAIL);
+
+        if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Starting with partial=%ld full=%ld verbose=%ld lowmemory=%ld MB vs freemem=%ld MB", fts_xapian_settings.partial,fts_xapian_settings.full,fts_xapian_settings.verbose,fts_xapian_settings.lowmemory, long(fts_backend_xapian_get_free_memory()/1024.0));
 
 	return 0;
 }
@@ -100,7 +118,7 @@ static void fts_backend_xapian_deinit(struct fts_backend *_backend)
 {
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *)_backend;
 
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Deinit %s)",backend->path);
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Deinit %s)",backend->path);
 
 	if(backend->guid != NULL) fts_backend_xapian_unset_box(backend);
 
@@ -114,12 +132,14 @@ static void fts_backend_xapian_deinit(struct fts_backend *_backend)
 	backend->path = NULL;
 
 	i_free(backend);
+
+	closelog();
 }
 
 
 static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend, struct mailbox *box, uint32_t *last_uid_r)
 {
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_get_last_uid");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_get_last_uid");
 
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *)_backend;
 
@@ -132,10 +152,9 @@ static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend, struct
 	}
 
 	Xapian::Database * dbr;
-
 	if(!fts_backend_xapian_open_readonly(backend, &dbr))
 	{
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: GetLastUID: Can not open db RO (%s)",backend->db);
+		i_error("FTS Xapian: GetLastUID: Can not open db RO (%s)",backend->db);
 		return 0;
 	}
 
@@ -145,17 +164,13 @@ static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend, struct
 	}
 	catch(Xapian::Error e)
 	{
-		i_error("FTS Xapian: fts_backend_xapian_get_last_uid for '%s' (%s)",backend->boxname,backend->guid);
-		i_error("FTS Xapian: %s",e.get_msg().c_str());
-		dbr->close();
-		delete(dbr);
-		return -1;
+		i_warning("FTS Xapian: fts_backend_xapian_get_last_uid for '%s' (%s)",backend->boxname,backend->guid);
+		i_warning("FTS Xapian: %s",e.get_msg().c_str());
 	}
 
 	dbr->close();
 	delete(dbr);
-
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Get last UID of %s (%s) = %d",backend->boxname,backend->guid,*last_uid_r);
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Get last UID of %s (%s) = %d",backend->boxname,backend->guid,*last_uid_r);
 
 	return 0;
 }
@@ -163,7 +178,7 @@ static int fts_backend_xapian_get_last_uid(struct fts_backend *_backend, struct
 
 static struct fts_backend_update_context * fts_backend_xapian_update_init(struct fts_backend *_backend)
 {
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_update_context");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_update_context");
 
 	struct xapian_fts_backend_update_context *ctx;
 
@@ -177,9 +192,7 @@ static int fts_backend_xapian_update_deinit(struct fts_backend_update_context *_
 	struct xapian_fts_backend_update_context *ctx = (struct xapian_fts_backend_update_context *)_ctx;
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *)ctx->ctx.backend;
 
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_deinit (%s)",backend->path);
-
-	fts_backend_xapian_release(backend,"update_deinit",0);
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_deinit (%s)",backend->path);
 
 	i_free(ctx);
 
@@ -188,7 +201,7 @@ static int fts_backend_xapian_update_deinit(struct fts_backend_update_context *_
 
 static void fts_backend_xapian_update_set_mailbox(struct fts_backend_update_context *_ctx, struct mailbox *box)
 {
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_set_mailbox");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_set_mailbox");
 
 	struct xapian_fts_backend_update_context *ctx = (struct xapian_fts_backend_update_context *)_ctx;
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *)ctx->ctx.backend;
@@ -198,7 +211,7 @@ static void fts_backend_xapian_update_set_mailbox(struct fts_backend_update_cont
 
 static void fts_backend_xapian_update_expunge(struct fts_backend_update_context *_ctx, uint32_t uid)
 {
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_update_expunge");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_expunge");
 
 	struct xapian_fts_backend_update_context *ctx = (struct xapian_fts_backend_update_context *)_ctx;
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *)ctx->ctx.backend;
@@ -220,6 +233,7 @@ static void fts_backend_xapian_update_expunge(struct fts_backend_update_context
 	else
 	{
 		char * u = i_strdup_printf("replace into docs values (%d)",uid);
+		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian : Expunged %d on %s",uid,backend->expdb);
 		if(sqlite3_exec(expdb,u,NULL,0,&zErrMsg) != SQLITE_OK)
 		{
 			i_error("FTS Xapian: Expunging (3) UID=%d : Can not add UID : %s",uid,zErrMsg);
@@ -228,6 +242,7 @@ static void fts_backend_xapian_update_expunge(struct fts_backend_update_context
 		i_free(u);
 	}
 	sqlite3_close(expdb);
+	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian : Expunge done");
 }
 
 static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_context *_ctx, const struct fts_backend_build_key *key)
@@ -252,29 +267,8 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 		fts_backend_xapian_oldbox(backend);
 		backend->old_guid = i_strdup(backend->guid);
 		backend->old_boxname = i_strdup(backend->boxname);
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Start indexing '%s' (%s)",backend->boxname,backend->guid);
 	}
 
-	/* Performance calculator*/
-	if( backend->perf_uid != key->uid )
-	{
-		backend->perf_nb++;
-		backend->perf_uid = key->uid;
-	}
-	if((backend->perf_nb - backend->perf_pt)>=200)
-	{
-		backend->perf_pt = backend->perf_nb;
-		long dt = fts_backend_xapian_current_time() - backend->perf_dt;
-		double r=0;
-		if(dt>0)
-		{
-			r=backend->perf_nb*1000.0;
-			r=r/dt;
-		}
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Partial indexing '%s' (%ld msgs in %ld ms, rate: %.1f)",backend->boxname,backend->perf_nb,dt,r);
-	}
-	/* End Performance calculator*/
-
 	const char * field=key->hdr_name;
 	const char * type = key->body_content_type;
 	const char * disposition = key->body_content_disposition;
@@ -285,13 +279,13 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 
 	if(key->type == FTS_BACKEND_BUILD_KEY_BODY_PART_BINARY)
 	{
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Skipping binary part of type '%s'",type);
+		if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Skipping binary part of type '%s'",type);
 		return FALSE;
 	}
 
 	if((type != NULL) && (strncmp(type,"text",4)!=0) && ((disposition==NULL) || ((strstr(disposition,"filename=")==NULL) && (strstr(disposition,"attachment")==NULL))))
 	{
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Non-binary & non-text part of type '%s'",type);
+		if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Non-binary & non-text part of type '%s'",type);
 		return FALSE;
 	}
 
@@ -299,7 +293,7 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 	ctx->isattachment=false;
 	if((disposition != NULL) && ((strstr(disposition,"filename=")!=NULL) || (strstr(disposition,"attachment")!=NULL)))
 	{
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Found part as attachment of type '%s' and disposition '%s'",type,disposition);
+		if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Found part as attachment of type '%s' and disposition '%s'",type,disposition);
 		ctx->isattachment=true;
 	}
 
@@ -345,9 +339,88 @@ static bool fts_backend_xapian_update_set_build_key(struct fts_backend_update_co
 			ctx->tbi_uid=key->uid;
 			break;
 		case FTS_BACKEND_BUILD_KEY_BODY_PART_BINARY:
-			i_unreached();
+		default:
+			return FALSE;
 	}
 
+	if((ctx->tbi_uid>0) && (ctx->tbi_uid != backend->lastuid))
+        {
+		long fri = fts_backend_xapian_test_memory();
+		if(backend->dbw == NULL)
+		{
+			bool ok=false;
+			long t=0,dt = fts_backend_xapian_current_time();
+			while(!ok)
+                        {
+                                try
+                                {
+                                        t++;
+                                        if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Opening (%s, %s) : Attempt %ld",backend->boxname,backend->db,t);
+                                        backend->dbw = new Xapian::WritableDatabase(backend->db,Xapian::DB_CREATE_OR_OPEN | Xapian::DB_BACKEND_GLASS);
+                                        ok=true;
+                                }
+                                catch(Xapian::DatabaseLockError e)
+                                {
+                                        if((t-1) % 20 ==0) 
+					{
+						long delay = (fts_backend_xapian_current_time() -dt)/1000.0;
+						if(delay > XAPIAN_MAX_SEC_WAIT) 
+						{
+							i_warning("FTS Xapian: Can't lock the DB (%s,%s) for %ld sec : Will try later",backend->boxname,backend->db,delay);
+							return FALSE;
+						}
+						if(fts_xapian_settings.verbose>0) i_warning("FTS Xapian: Can't lock the DB (%s,%s) for %ld sec : %s - %s %s",backend->boxname,backend->db,delay,e.get_type(),e.get_msg().c_str(),e.get_error_string());
+					}
+                                        std::this_thread::sleep_for(XSLEEP);
+                                }
+                                catch(Xapian::Error e)
+                                {
+                                        i_error("FTS Xapian: Can't open Xapian DB %s 2 : %s - %s %s ",backend->db,e.get_type(),e.get_msg().c_str(),e.get_error_string());
+                                        return false;
+                                }
+                        }
+			dt=fts_backend_xapian_current_time() - dt;
+			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: DB (%s,%s) opened ! in %ld ms",backend->boxname,backend->db,dt);
+		}
+		if(backend->lastuid>0)
+		{
+			std::string s("New doc ready to index "); s.append(std::to_string(backend->lastuid));
+			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: %s",s.c_str());
+			while((!fts_backend_xapian_push(backend,s.c_str())) && ((backend->docs->size()>XAPIAN_THREAD_SIZE*backend->threads_max) || (fri>=0)))
+			{
+				if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Waiting for an available thread (Sleep1) : Pending are %ld",backend->docs->size());
+				std::this_thread::sleep_for(XSLEEP);
+			}
+		}
+		if((fri>=0) && (backend->dbw != NULL))
+        	{
+                	i_warning("FTS Xapian: Warning Free memory %ld MB < %ld MB minimum, Writing to disk (creating lock)",long(fri/1024.0),fts_xapian_settings.lowmemory);
+			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Low memory, Writing to disk (creating lock)");
+                                
+			std::unique_lock<std::timed_mutex> lck(backend->mutex,std::defer_lock);
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
+                        while(!(lck.try_lock_for(std::chrono::milliseconds(1000 + std::rand() % 1000))))
+                        {
+                                if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: Waiting unlock... (main)");
+                        }
+#pragma GCC diagnostic pop
+                        if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Lock acquired (main)");
+			try
+			{
+				backend->dbw->commit();
+			}
+			catch(Xapian::Error e)
+			{
+				i_error("FTS Xapian: Sync to disk, error committing db %s",e.get_msg().c_str());
+			}
+			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Sync to disc done");
+        	}
+                backend->lastuid = ctx->tbi_uid;
+		backend->docs->push_back(new XDoc(backend->lastuid));
+		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Start indexing #%ld (%s) : Buffer %ld",backend->lastuid, backend->boxname,backend->docs->size());
+        }
+
 	return TRUE;
 }
 
@@ -355,24 +428,7 @@ static void fts_backend_xapian_update_unset_build_key(struct fts_backend_update_
 {
 	struct xapian_fts_backend_update_context *ctx = (struct xapian_fts_backend_update_context *)_ctx;
 
-	if(fts_xapian_settings.verbose>0) 
-	{
-		long n = 0;
-		if( ctx->ctx.backend != NULL )
-		{
-			struct xapian_fts_backend *backend = (struct xapian_fts_backend *)ctx->ctx.backend;
-			if( backend->dbw != NULL) n = backend->dbw->get_doccount();
-		}
-
-		if(n>0)
-		{
-			i_info("FTS Xapian: fts_backend_xapian_update_unset_build_key with %ld docs in the index",n);
-		}
-		else
-		{
-			i_info("FTS Xapian: fts_backend_xapian_update_unset_build_key");
-		}
-	}
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_update_unset_build_key");
 
 	if(ctx->tbi_field!=NULL)
 	{
@@ -384,12 +440,10 @@ static void fts_backend_xapian_update_unset_build_key(struct fts_backend_update_
 
 static int fts_backend_xapian_refresh(struct fts_backend * _backend)
 {
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_refresh");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_refresh");
 
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *) _backend;
 
-	fts_backend_xapian_release(backend,"refresh", 0);
-
 	return 0;
 }
 
@@ -398,94 +452,19 @@ static int fts_backend_xapian_update_build_more(struct fts_backend_update_contex
 	struct xapian_fts_backend_update_context *ctx = (struct xapian_fts_backend_update_context *)_ctx;
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *) ctx->ctx.backend;
 
-	if(fts_xapian_settings.verbose>1)
-	{
-		if(ctx->isattachment)
-		{
-			i_info("FTS Xapian: Indexing part as attachment");
-		}
-		else
-		{
-			i_info("FTS Xapian: Indexing part as text");
-		}
-	}
-
 	if(ctx->tbi_uid<1) return 0;
 
 	if(data == NULL) return 0;
 
 	icu::StringPiece sp_d((const char *)data,(int32_t )size);
 	icu::UnicodeString d2 = icu::UnicodeString::fromUTF8(sp_d);
+	
 	if(d2.length() < fts_xapian_settings.partial) return 0;
 
-	if(!fts_backend_xapian_check_access(backend))
-	{
-		i_error("FTS Xapian: Buildmore: Can not open db");
-		return -1;
-	}
-
-	long fri = fts_backend_xapian_test_memory();
-
-	if(fri>=0)
-	{
-		i_warning("FTS Xapian: Warning Free memory %ld MB < %ld MB minimum",long(fri/1024.0),fts_xapian_settings.lowmemory);
-		fts_backend_xapian_release(backend,"Low memory indexing", 0);
-		if(!fts_backend_xapian_check_access(backend))
-		{
-			i_error("FTS Xapian: Buildmore: Can not open db (2)");
-			return -1;
-		}
-	}
-
-	bool ok=true;
-
-	if(ctx->tbi_isfield)
-	{
-		ok=fts_backend_xapian_index_hdr(backend,ctx->tbi_uid,ctx->tbi_field, &d2);
-		if(!ok)
-		{
-			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Flushing memory and retrying");
-			fts_backend_xapian_release(backend,"Flushing memory indexing hdr", 0);
-                	if(fts_backend_xapian_check_access(backend))
-			{
-				ok=fts_backend_xapian_index_hdr(backend,ctx->tbi_uid,ctx->tbi_field, &d2);
-			}
-			else
-                	{
-                        	i_error("FTS Xapian: Buildmore: Can not open db (3)");
-                	}
-		}
-	}
-	else
-	{
-		ok=fts_backend_xapian_index_text(backend,ctx->tbi_uid,ctx->tbi_field, &d2);
-		if(!ok)
-		{
-			if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Flushing memory and retrying");
-			fts_backend_xapian_release(backend,"Flushing memory indexing text", 0);
-			if(fts_backend_xapian_check_access(backend))
-			{
-				ok=fts_backend_xapian_index_text(backend,ctx->tbi_uid,ctx->tbi_field, &d2);
-			}
-			else
-			{
-				i_error("FTS Xapian: Buildmore: Can not open db (4)");
-			}
-		}
-	}
-
-	backend->commit_updates++;
-
-	long current_time = fts_backend_xapian_current_time();
-
-	if( (!ok) || (backend->commit_updates>XAPIAN_COMMIT_ENTRIES) || ((current_time - backend->commit_time) > XAPIAN_COMMIT_TIMEOUT*1000) )
-	{
-		if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Refreshing after %ld ms (vs %ld) and %ld updates (vs %ld) ...", current_time - backend->commit_time, XAPIAN_COMMIT_TIMEOUT*1000, backend->commit_updates, XAPIAN_COMMIT_ENTRIES);
-		fts_backend_xapian_release(backend,"refreshing", current_time);
-	}
-
-	if(!ok) return -1;
-	return 0;
+	if(fts_backend_xapian_index(backend,ctx->tbi_field, &d2)) return 0;
+        	
+	i_error("FTS Xapian: Buildmore: Error Index");
+	return -1;
 }
 
 static int fts_backend_xapian_optimize_callback(void *data, int argc, char **argv, char **azColName)
@@ -496,7 +475,7 @@ static int fts_backend_xapian_optimize_callback(void *data, int argc, char **arg
 		return -1;
 	}
 	uint32_t uid = atol(argv[0]);
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_optimize_callback : Adding %d",uid);
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_optimize_callback : Adding %d",uid);
 	std::vector<uint32_t> * uids = (std::vector<uint32_t> *) data;
 	uids->push_back(uid);
 	return 0;
@@ -506,12 +485,12 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 {
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *) _backend;
 
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_optimize '%s'",backend->path);
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_optimize '%s'",backend->path);
 
 	struct stat sb;
 	if(!( (stat(backend->path, &sb)==0) && S_ISDIR(sb.st_mode)))
 	{
-		if(fts_xapian_settings.verbose>0) i_error("FTS Xapian: Optimize(0) Index folder inexistent");
+		i_error("FTS Xapian: Optimize(0) Index folder inexistent");
 		return -1;
 	}
 
@@ -526,17 +505,20 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 	char *zErrMsg = 0;
 	XResultSet * result = NULL;
 	Xapian::docid docid =0;
-
+	struct stat fileinfo;
 	while ((dp = readdir(dirp)) != NULL)
 	{
 		if((dp->d_type == DT_DIR) && (strncmp(dp->d_name,"db_",3)==0))
 		{
 			uids.clear();
+			s = i_strdup_printf("%s/%s",backend->path,dp->d_name);
+			stat(s, &fileinfo);
+			i_free(s);
 			s = i_strdup_printf("%s/%s_exp.db",backend->path,dp->d_name);
-			if(fts_xapian_settings.verbose>0) i_info("Optimize (1) %s : Checking expunges",s);
+			i_info("FTS Xapian: Optimize (1) %s : Checking expunges",s);
 			if(sqlite3_open(s,&expdb) == SQLITE_OK)
 			{
-				if(fts_xapian_settings.verbose>0) i_info("Optimize (1b) Executing %s",createTable);
+				if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize (1b) Executing %s",createTable);
 				if(sqlite3_exec(expdb,createTable,NULL,0,&zErrMsg) != SQLITE_OK )
 				{
 					i_error("FTS Xapian: Optimize (2) Can not create table (%s) : %s",createTable,zErrMsg);
@@ -545,7 +527,7 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 				}
 				else
 				{
-					if(fts_xapian_settings.verbose>0) i_info("Optimize (1c) Executing %s",selectUIDs);
+					if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize (1c) Executing %s",selectUIDs);
 					if(sqlite3_exec(expdb,selectUIDs,fts_backend_xapian_optimize_callback,&uids,&zErrMsg) != SQLITE_OK)	
 					{
 						i_error("FTS Xapian: Optimize (3) Can not select IDs (%s) : %s",selectUIDs,zErrMsg);
@@ -555,10 +537,24 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 				}
 				i_free(s);
 				s = i_strdup_printf("%s/%s",backend->path,dp->d_name);
-				if(fts_xapian_settings.verbose>0) i_info("Optimize (4) Opening Xapian DB (%s)",s);
+				if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize (4) Opening Xapian DB (%s)",s);
 				try
 				{
-					db = new Xapian::WritableDatabase(s,Xapian::DB_CREATE_OR_OPEN | Xapian::DB_RETRY_LOCK | Xapian::DB_BACKEND_GLASS | Xapian::DB_NO_SYNC);
+					bool ok=false;
+					while(!ok)
+					{
+						try
+                                		{
+							db = new Xapian::WritableDatabase(s,Xapian::DB_CREATE_OR_OPEN | Xapian::DB_BACKEND_GLASS);
+                                        		ok=true;
+                                		}
+                                		catch(Xapian::DatabaseLockError e)
+                                		{
+                                		        i_warning("FTS Xapian: Retrying opening DB %s - %s %s",e.get_type(),e.get_msg().c_str(),e.get_error_string());
+                                        		std::this_thread::sleep_for(XSLEEP);
+						}
+                                	}
+					long c=0;
 					for(uint32_t n=0;n<uids.size();n++)
 					{
 						uid=uids[n];
@@ -568,7 +564,7 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 						xq->add("uid",u);
 						i_free(u);
 						result=fts_backend_xapian_query(db,xq,1);
-						docid=0;
+						docid=0; 
 						if((result!=NULL) && (result->size>0))
 						{
 							try
@@ -576,6 +572,13 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 								docid = result->data[0];
 								if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize (5) Removing DOC UID=%d DOCID=%d",uid,docid);
 								db->delete_document(docid);
+								c++;
+								if(c>XAPIAN_WRITING_CACHE) 
+								{
+									i_info("FTS Xapian: Flushing changes on disk");
+									db->commit();
+									c=0;
+								}
 							}
 							catch(Xapian::Error e)
 							{
@@ -593,13 +596,23 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 						}
 						i_free(u);
 					}
-					db->commit();
-					db->close();
-					delete(db);
+					if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: Optimize - Closing DB %s",s);
+					char * s1 = (char *)malloc(sizeof(char) * (strlen(s)+1));
+					strcpy(s1,s);
+					char * s2 = (char *)malloc(sizeof(char) * 13);
+					strcpy(s2,"fts_optimize");
+					if(fts_xapian_settings.detach)
+					{
+						(new std::thread(fts_backend_xapian_close_db,db,s1,s2,fileinfo.st_uid,fileinfo.st_gid,fts_xapian_settings.verbose,true))->detach();
+					}
+					else
+					{
+						fts_backend_xapian_close_db(db,s1,s2,fileinfo.st_uid,fileinfo.st_gid,fts_xapian_settings.verbose,false);
+					}
 				}
 				catch(Xapian::Error e)
 				{
-					i_error("FTS Xapian: Optimize (6) %s",e.get_msg().c_str());
+					i_error("FTS Xapian: Optimize (7) %s",e.get_msg().c_str());
 				}
 				sqlite3_close(expdb);
 			}
@@ -612,7 +625,7 @@ static int fts_backend_xapian_optimize(struct fts_backend *_backend)
 
 static int fts_backend_xapian_rescan(struct fts_backend *_backend)
 {
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_rescan");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_rescan");
 
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *) _backend;
 
@@ -632,7 +645,7 @@ static int fts_backend_xapian_rescan(struct fts_backend *_backend)
 
 		if((dp->d_type == DT_REG) && (strncmp(dp->d_name,"expunge_",8)==0))
 		{
-			if(fts_xapian_settings.verbose>0) i_info("Removing[1] %s",s);
+			i_info("Removing[1] %s",s);
 			remove(s);
 		}
 		else if((dp->d_type == DT_DIR) && (strncmp(dp->d_name,"db_",3)==0))
@@ -644,13 +657,13 @@ static int fts_backend_xapian_rescan(struct fts_backend *_backend)
 				s2 = i_strdup_printf("%s/%s",s,dp2->d_name);
 				if(dp2->d_type == DT_REG)
 				{
-					if(fts_xapian_settings.verbose>0) i_info("Removing[2] %s",s2);
+					i_info("Removing[2] %s",s2);
 					remove(s2);
 				}
 				i_free(s2);
 			}
 			closedir(d2);
-			if(fts_xapian_settings.verbose>0) i_info("Removing dir %s",s);
+			i_info("Removing dir %s",s);
 			remove(s);
 		}
 		i_free(s);
@@ -662,7 +675,7 @@ static int fts_backend_xapian_rescan(struct fts_backend *_backend)
 
 static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbox *box, struct mail_search_arg *args, enum fts_lookup_flags flags, struct fts_result *result)
 {
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_lookup");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_lookup");
 
 	struct xapian_fts_backend *backend = (struct xapian_fts_backend *) _backend;
 
@@ -699,7 +712,7 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
 	XResultSet * r=fts_backend_xapian_query(dbr,qs);
 
 	long n=r->size;
-	if(fts_xapian_settings.verbose>0) { i_info("FTS Xapian: QUery '%s' -> %ld results",qs->get_string().c_str(),n); }
+	if(fts_xapian_settings.verbose>0) { i_info("FTS Xapian: Query '%s' -> %ld results",qs->get_string().c_str(),n); }
 
 	i_array_init(&(result->definite_uids),r->size);
 
@@ -730,7 +743,7 @@ static int fts_backend_xapian_lookup(struct fts_backend *_backend, struct mailbo
 
 static int fts_backend_xapian_lookup_multi(struct fts_backend *_backend, struct mailbox *const boxes[], struct mail_search_arg *args, enum fts_lookup_flags flags, struct fts_multi_result *result)
 {
-	if(fts_xapian_settings.verbose>0) i_info("FTS Xapian: fts_backend_xapian_lookup_multi");
+	if(fts_xapian_settings.verbose>1) i_info("FTS Xapian: fts_backend_xapian_lookup_multi");
 
 	ARRAY(struct fts_result) box_results;
 
